import numpy as np
from numpy.linalg import lstsq

from amodem import dsp
from amodem import sampling

import itertools
import random

_constellation = [1, 1j, -1, -1j]


class Equalizer(object):

    def __init__(self, config):
        self.carriers = config.carriers
        self.omegas = 2 * np.pi * np.array(config.frequencies) / config.Fs
        self.Nfreq = config.Nfreq
        self.Nsym = config.Nsym

    def train_symbols(self, length, seed=0):
        r = random.Random(seed)
        choose = lambda: [r.choice(_constellation) for j in range(self.Nfreq)]
        return np.array([choose() for i in range(length)])

    def modulator(self, symbols):
        gain = 1.0 / len(self.carriers)
        result = []
        for s in symbols:
            result.append(np.dot(s, self.carriers))
        result = np.concatenate(result).real * gain
        assert np.max(np.abs(result)) <= 1
        return result

    def demodulator(self, signal, size):
        signal = itertools.chain(signal, itertools.repeat(0))
        symbols = dsp.Demux(sampler=sampling.Sampler(signal),
                            omegas=self.omegas, Nsym=self.Nsym)
        return np.array(list(itertools.islice(symbols, size)))

    def equalize_symbols(self, signal, symbols, order, lookahead=0):
        assert symbols.shape[1] == self.Nfreq
        length = symbols.shape[0]

        matched = np.array(self.carriers) / (0.5*self.Nsym)
        matched = matched[:, ::-1].transpose().conj()
        signal = np.concatenate([signal, np.zeros(lookahead)])
        y = dsp.lfilter(x=signal, b=matched, a=[1])

        A = []
        b = []

        for j in range(self.Nfreq):
            for i in range(length):
                offset = (i+1)*self.Nsym
                row = y[offset-order:offset+lookahead, j]
                A.append(row)
                b.append(symbols[i, j])

        A = np.array(A)
        b = np.array(b)
        h, residuals, rank, sv = lstsq(A, b)
        h = h[::-1].real

        return h

    def equalize_signal(self, signal, expected, order, lookahead=0):
        signal = [np.zeros(order-1), signal, np.zeros(lookahead)]
        signal = np.concatenate(signal)
        length = len(expected)

        A = []
        b = []

        for i in range(length - order):
            offset = order + i
            row = signal[offset-order:offset+lookahead]
            A.append(np.array(row, ndmin=2))
            b.append(expected[i])

        A = np.concatenate(A, axis=0)
        b = np.array(b)
        h, residuals, rank, sv = lstsq(A, b)
        h = h[::-1].real
        return h



import numpy as np
import logging
import itertools
import functools
import time

log = logging.getLogger(__name__)

from . import stream
from . import dsp
from . import sampling
from . import train
from . import common
from . import framing
from . import equalizer
from . import detect


class Receiver(object):

    def __init__(self, config, pylab=None):
        self.stats = {}
        self.plt = pylab or common.Dummy()
        self.modem = dsp.MODEM(config.symbols)
        self.frequencies = np.array(config.frequencies)
        self.omegas = 2 * np.pi * self.frequencies / config.Fs
        self.Nsym = config.Nsym
        self.Tsym = config.Tsym
        self.iters_per_report = config.baud  # report once per second
        self.modem_bitrate = config.modem_bps
        self.equalizer = equalizer.Equalizer(config)
        self.carrier_index = config.carrier_index

    def _prefix(self, symbols, gain=1.0, skip=5):
        S = common.take(symbols, len(train.prefix))
        S = S[:, self.carrier_index] * gain
        sliced = np.round(np.abs(S))
        self.plt.figure()
        self.plt.subplot(121)
        self._constellation(S, sliced, 'Prefix')

        bits = np.array(sliced, dtype=int)
        self.plt.subplot(122)
        self.plt.plot(np.abs(S))
        self.plt.plot(train.prefix)
        if any(bits != train.prefix):
            raise ValueError('Incorrect prefix')

        log.debug('Prefix OK')

        nonzeros = np.array(train.prefix, dtype=bool)
        pilot_tone = S[nonzeros]
        phase = np.unwrap(np.angle(pilot_tone)) / (2 * np.pi)
        indices = np.arange(len(phase))
        a, b = dsp.linear_regression(indices[skip:-skip], phase[skip:-skip])
        self.plt.figure()
        self.plt.plot(indices, phase, ':')
        self.plt.plot(indices, a * indices + b)

        freq_err = a / (self.Tsym * self.frequencies[self.carrier_index])
        last_phase = a * indices[-1] + b
        log.debug('Current phase on carrier: %.3f', last_phase)

        log.debug('Frequency error: %.2f ppm', freq_err * 1e6)
        self.plt.title('Frequency drift: {0:.3f} ppm'.format(freq_err * 1e6))
        return freq_err

    def _train(self, sampler, order, lookahead):
        Nfreq = len(self.frequencies)
        train_symbols = self.equalizer.train_symbols(train.equalizer_length)
        train_signal = self.equalizer.modulator(train_symbols) * Nfreq

        prefix = postfix = train.silence_length * self.Nsym
        signal_length = train.equalizer_length * self.Nsym + prefix + postfix

        signal = sampler.take(signal_length + lookahead)

        coeffs = self.equalizer.equalize_signal(
            signal=signal[prefix:-postfix],
            expected=train_signal,
            order=order, lookahead=lookahead
        )

        self.plt.figure()
        self.plt.plot(np.arange(order+lookahead), coeffs)

        equalization_filter = dsp.FIR(h=coeffs)
        equalized = list(equalization_filter(signal))
        equalized = equalized[prefix+lookahead:-postfix+lookahead]

        symbols = self.equalizer.demodulator(equalized, train.equalizer_length)
        sliced = np.array(symbols).round()
        errors = np.array(sliced - train_symbols, dtype=np.bool)
        error_rate = errors.sum() / errors.size

        errors = np.array(symbols - train_symbols)
        rms = lambda x: (np.mean(np.abs(x) ** 2, axis=0) ** 0.5)

        noise_rms = rms(errors)
        signal_rms = rms(train_symbols)
        SNRs = 20.0 * np.log10(signal_rms / noise_rms)

        self.plt.figure()
        for (i, freq), snr in zip(enumerate(self.frequencies), SNRs):
            log.debug('%5.1f kHz: SNR = %5.2f dB', freq / 1e3, snr)
            self._constellation(symbols[:, i], train_symbols[:, i],
                                '$F_c = {0} Hz$'.format(freq), index=i)
        assert error_rate == 0, error_rate

        return equalization_filter

    def _demodulate(self, sampler, symbols):
        streams = []
        symbol_list = []
        errors = {}

        def error_handler(received, decoded, freq):
            errors.setdefault(freq, []).append(received / decoded)

        generators = common.split(symbols, n=len(self.omegas))
        for freq, S in zip(self.frequencies, generators):
            equalized = []
            S = common.icapture(S, result=equalized)
            symbol_list.append(equalized)

            freq_handler = functools.partial(error_handler, freq=freq)
            bits = self.modem.decode(S, freq_handler)  # list of bit tuples
            streams.append(bits)  # stream per frequency

        self.stats['symbol_list'] = symbol_list
        self.stats['rx_bits'] = 0
        self.stats['rx_start'] = time.time()

        log.info('Starting demodulation')
        for i, block in enumerate(common.izip(streams), 1):
            for bits in block:
                self.stats['rx_bits'] = self.stats['rx_bits'] + len(bits)
                yield bits

            if i % self.iters_per_report == 0:
                err = np.array([e for v in errors.values() for e in v])
                err = np.mean(np.angle(err))/(2*np.pi) if len(err) else 0
                errors.clear()

                duration = time.time() - self.stats['rx_start']
                sampler.freq -= 0.01 * err * self.Tsym
                sampler.offset -= err
                log.debug(
                    'Got  %8.1f kB, realtime: %6.2f%%, drift: %+5.2f ppm',
                    self.stats['rx_bits'] / 8e3,
                    duration * 100.0 / (i*self.Tsym),
                    (1.0 - sampler.freq) * 1e6
                )

    def start(self, signal, gain=1.0):
        sampler = sampling.Sampler(signal, sampling.Interpolator())
        symbols = dsp.Demux(sampler, omegas=self.omegas, Nsym=self.Nsym)
        freq_err = self._prefix(symbols, gain=gain)
        sampler.freq -= freq_err

        filt = self._train(sampler, order=11, lookahead=6)
        sampler.equalizer = lambda x: list(filt(x))

        bitstream = self._demodulate(sampler, symbols)
        self.bitstream = itertools.chain.from_iterable(bitstream)

    def run(self, output):
        data = framing.decode(self.bitstream)
        self.size = 0
        for chunk in common.iterate(data=data, size=256,
                                    truncate=False, func=bytearray):
            output.write(chunk)
            self.size += len(chunk)

    def report(self):
        if self.stats:
            duration = time.time() - self.stats['rx_start']
            audio_time = self.stats['rx_bits'] / float(self.modem_bitrate)
            log.debug('Demodulated %.3f kB @ %.3f seconds (%.1f%% realtime)',
                      self.stats['rx_bits'] / 8e3, duration,
                      100 * duration / audio_time if audio_time else 0)

            log.info('Received %.3f kB @ %.3f seconds = %.3f kB/s',
                     self.size * 1e-3, duration, self.size * 1e-3 / duration)

            self.plt.figure()
            symbol_list = np.array(self.stats['symbol_list'])
            for i, freq in enumerate(self.frequencies):
                self._constellation(symbol_list[i], self.modem.symbols,
                                    '$F_c = {0} Hz$'.format(freq), index=i)
        self.plt.show()

    def _constellation(self, y, symbols, title, index=None):
        if index is not None:
            Nfreq = len(self.frequencies)
            height = np.floor(np.sqrt(Nfreq))
            width = np.ceil(Nfreq / float(height))
            self.plt.subplot(height, width, index + 1)

        theta = np.linspace(0, 2*np.pi, 1000)
        y = np.array(y)
        self.plt.plot(y.real, y.imag, '.')
        self.plt.plot(np.cos(theta), np.sin(theta), ':')
        points = np.array(symbols)
        self.plt.plot(points.real, points.imag, '+')
        self.plt.grid('on')
        self.plt.axis('equal')
        self.plt.axis(np.array([-1, 1, -1, 1])*1.1)
        self.plt.title(title)


def main(config, src, dst, pylab=None):
    reader = stream.Reader(src, data_type=common.loads)
    signal = itertools.chain.from_iterable(reader)

    to_skip = int(config.skip_start * config.Fs)
    log.debug('Skipping %.3f seconds', config.skip_start)
    common.take(signal, to_skip)

    reader.check = common.check_saturation

    detector = detect.Detector(config=config)
    receiver = Receiver(config=config, pylab=pylab)
    success = False
    try:
        log.info('Waiting for carrier tone: %.1f kHz', config.Fc / 1e3)
        signal, amplitude = detector.run(signal)
        receiver.start(signal, gain=1.0/amplitude)
        receiver.run(dst)
        success = True
    except Exception:
        log.exception('Decoding failed')

    receiver.report()
    return success



import time
import logging

log = logging.getLogger(__name__)


class Reader(object):

    def __init__(self, fd, data_type=None, bufsize=4096,
                 eof=False, timeout=2.0, wait=0.2):
        self.fd = fd
        self.data_type = data_type if (data_type is not None) else lambda x: x
        self.bufsize = bufsize
        self.eof = eof
        self.timeout = timeout
        self.wait = wait
        self.total = 0
        self.check = None

    def __iter__(self):
        return self

    def next(self):
        block = bytearray()
        if self.eof:
            data = self.fd.read(self.bufsize)
            if data:
                self.total += len(data)
                block.extend(data)
                return block
            else:
                raise StopIteration()

        finish_time = time.time() + self.timeout
        while time.time() <= finish_time:
            left = self.bufsize - len(block)
            data = self.fd.read(left)
            if data:
                self.total += len(data)
                block.extend(data)

            if len(block) == self.bufsize:
                values = self.data_type(block)
                if self.check:
                    self.check(values)
                return values

            time.sleep(self.wait)

        raise IOError('timeout')

    __next__ = next



import numpy as np
import logging

log = logging.getLogger(__name__)

from . import common

ALLOWED_EXCEPTIONS = (IOError, KeyboardInterrupt)


def send(config, dst, src=None, verbose=False):
    calibration_symbols = int(1.0 * config.Fs)
    t = np.arange(0, calibration_symbols) * config.Ts
    signal = [np.sin(2 * np.pi * f * t) for f in config.frequencies]
    signal = common.dumps(np.concatenate(signal))

    try:
        while True:
            dst.write(signal)
    except ALLOWED_EXCEPTIONS:
        pass


def run_recorder(config, recorder):

    frame_length = 200 * config.Nsym
    frame_size = frame_length * config.sample_size

    t = np.arange(0, frame_length) * config.Ts

    scaling_factor = 0.5 * len(t)
    carriers = [np.exp(2j * np.pi * f * t) for f in config.frequencies]
    carriers = np.array(carriers) / scaling_factor

    states = [True]
    errors = ['weak', 'strong', 'noisy']
    try:
        while True:
            data = recorder.read(frame_size)
            if len(data) < frame_size:
                return
            data = common.loads(data)
            frame = data - np.mean(data)

            coeffs = np.dot(carriers, frame)
            peak = np.max(np.abs(frame))
            total = np.sqrt(np.dot(frame, frame) / scaling_factor)

            max_index = np.argmax(np.abs(coeffs))
            freq = config.frequencies[max_index]
            rms = abs(coeffs[max_index])
            coherency = rms / total
            flags = [rms > 0.1, peak < 1.0, coherency > 0.99]

            states.append(all(flags))
            states = states[-2:]

            message = 'good signal'
            error = not any(states)
            if error:
                error_index = flags.index(False)
                message = 'too {0} signal'.format(errors[error_index])

            yield dict(
                freq=freq, rms=rms, peak=peak, coherency=coherency,
                total=total, error=error, message=message
            )
    except ALLOWED_EXCEPTIONS:
        pass

fmt = '{freq:6.0f} Hz: {message:s}{extra:s}'
fields = ['peak', 'total', 'rms', 'coherency']


def recv(config, src, dst=None, verbose=False):
    extra = ''
    if verbose:
        extra = ''.join(', {0}={{{0}:.4f}}'.format(f) for f in fields)

    for result in run_recorder(config=config, recorder=src):
        msg = fmt.format(extra=extra.format(**result), **result)
        if not result['error']:
            log.info(msg)
        else:
            log.error(msg)



import pyaudio
import logging

log = logging.getLogger(__name__)


class Interface(object):
    def __init__(self, config, library=pyaudio):
        self.p = library.PyAudio()
        format = getattr(library, 'paInt{0}'.format(config.bits_per_sample))
        self.sample_size = config.sample_size
        self.kwargs = dict(
            channels=1, rate=int(config.Fs), format=format,
            frames_per_buffer=config.samples_per_buffer
        )

    def __del__(self):
        self.p.terminate()

    def player(self):
        return self.p.open(output=True, **self.kwargs)

    def recorder(self):
        stream = self.p.open(input=True, **self.kwargs)
        return _Recorder(stream, sample_size=self.sample_size)


class _Recorder(object):
    def __init__(self, stream, sample_size):
        self.stream = stream
        self.sample_size = sample_size

    def read(self, size):
        assert size % self.sample_size == 0
        return self.stream.read(size // self.sample_size)



''' Reed-Solomon CODEC. '''
from . import common
import bitarray

import functools
import itertools
import binascii
import struct
import logging
log = logging.getLogger(__name__)

_crc32 = lambda x, mask: binascii.crc32(x) & mask
# (so the result will be unsigned on Python 2/3)


class Checksum(object):
    fmt = '>L'  # unsigned longs (32-bit)
    size = struct.calcsize(fmt)
    func = functools.partial(_crc32, mask=0xFFFFFFFF)

    def encode(self, payload):
        checksum = self.func(payload)
        return struct.pack(self.fmt, checksum) + payload

    def decode(self, data):
        received, = struct.unpack(self.fmt, data[:self.size])
        payload = data[self.size:]
        expected = self.func(payload)
        if received != expected:
            log.warning('Invalid checksum: %04x != %04x', received, expected)
            raise ValueError('Invalid checksum')
        return payload


class Framer(object):
    block_size = 1024
    prefix_fmt = '>L'
    prefix_len = struct.calcsize(prefix_fmt)
    checksum = Checksum()

    EOF = b''

    def _pack(self, block):
        frame = self.checksum.encode(block)
        return struct.pack(self.prefix_fmt, len(frame)) + frame

    def encode(self, data):
        for block in common.iterate(data=data, size=self.block_size,
                                    func=bytearray, truncate=False):
            yield self._pack(block=block)
        yield self._pack(block=self.EOF)

    def decode(self, data):
        data = iter(data)
        while True:
            length, = self._take_fmt(data, self.prefix_fmt)
            frame = self._take_len(data, length)
            block = self.checksum.decode(frame)
            if block == self.EOF:
                log.debug('EOF frame detected')
                return

            yield block

    def _take_fmt(self, data, fmt):
        length = struct.calcsize(fmt)
        chunk = bytearray(itertools.islice(data, length))
        if len(chunk) < length:
            raise ValueError('missing prefix data')
        return struct.unpack(fmt, chunk)

    def _take_len(self, data, length):
        chunk = bytearray(itertools.islice(data, length))
        if len(chunk) < length:
            raise ValueError('missing payload data')
        return chunk


def chain_wrapper(func):
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        result = func(*args, **kwargs)
        return itertools.chain.from_iterable(result)
    return wrapped


@chain_wrapper
def encode(data, framer=None):
    framer = framer or Framer()
    for frame in framer.encode(data):
        bits = bitarray.bitarray(endian='little')
        bits.frombytes(bytes(frame))
        yield bits


@chain_wrapper
def _to_bytes(bits, block_size=1):
    for chunk in common.iterate(data=bits, size=8*block_size,
                                func=lambda x: x, truncate=True):
        yield bitarray.bitarray(chunk, endian='little').tobytes()


@chain_wrapper
def decode(bits, framer=None):
    framer = framer or Framer()
    for frame in framer.decode(_to_bytes(bits)):
        yield frame



import numpy as np
import logging
import itertools
import collections

log = logging.getLogger(__name__)

from . import dsp
from . import train
from . import common


class Detector(object):

    COHERENCE_THRESHOLD = 0.9

    CARRIER_DURATION = sum(train.prefix)
    CARRIER_THRESHOLD = int(0.9 * CARRIER_DURATION)
    SEARCH_WINDOW = int(0.1 * CARRIER_DURATION)

    TIMEOUT = 10.0  # [seconds]

    def __init__(self, config):
        self.freq = config.Fc
        self.omega = 2 * np.pi * self.freq / config.Fs
        self.Nsym = config.Nsym
        self.Tsym = config.Tsym
        self.maxlen = config.baud  # 1 second of symbols
        self.max_offset = self.TIMEOUT * config.Fs

    def run(self, samples):
        counter = 0
        bufs = collections.deque([], maxlen=self.maxlen)
        for offset, buf in common.iterate(samples, self.Nsym, enumerate=True):
            if offset > self.max_offset:
                raise ValueError('Timeout waiting for carrier')
            bufs.append(buf)

            coeff = dsp.coherence(buf, self.omega)
            if abs(coeff) > self.COHERENCE_THRESHOLD:
                counter += 1
            else:
                counter = 0

            if counter == self.CARRIER_THRESHOLD:
                break
        else:
            raise ValueError('No carrier detected')

        length = (self.CARRIER_THRESHOLD - 1) * self.Nsym
        begin = offset - length

        x = np.concatenate(tuple(bufs)[-self.CARRIER_THRESHOLD:-1])
        Hc = dsp.exp_iwt(-self.omega, len(x))
        Zc = np.dot(Hc, x) / (0.5*len(x))
        amplitude = abs(Zc)
        start_time = begin * self.Tsym / self.Nsym
        log.info('Carrier detected at ~%.1f ms @ %.1f kHz:'
                 ' coherence=%.3f%%, amplitude=%.3f',
                 start_time * 1e3, self.freq / 1e3,
                 np.abs(dsp.coherence(x, self.omega)) * 100, amplitude)

        log.debug('Buffered %d ms of audio', len(bufs))

        bufs = list(bufs)[-self.CARRIER_THRESHOLD-self.SEARCH_WINDOW:]
        n = self.SEARCH_WINDOW + self.CARRIER_DURATION - self.CARRIER_THRESHOLD
        trailing = list(itertools.islice(samples, n * self.Nsym))
        bufs.append(np.array(trailing))

        buf = np.concatenate(bufs)
        offset = self.find_start(buf, self.CARRIER_DURATION*self.Nsym)
        start_time += (offset / self.Nsym - self.SEARCH_WINDOW) * self.Tsym
        log.debug('Carrier starts at %.3f ms', start_time * 1e3)

        return itertools.chain(buf[offset:], samples), amplitude

    def find_start(self, buf, length):
        N = len(buf)
        carrier = dsp.exp_iwt(self.omega, N)
        z = np.cumsum(buf * carrier)
        z = np.concatenate([[0], z])
        correlations = np.abs(z[length:] - z[:-length])
        return np.argmax(correlations)
